# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_static.ipynb.

# %% auto 0
__all__ = ['LOG_SAFETY', 'SSR_CHI2TEST', 'LAG_ORDER', 'DEFAULT_TEST', 'DEFAULT_GRANGER_CAUSALITY_TESTS', 'LOG2_FOLD',
           'SIGNED_CORRELATION', 'STANDARD_SCALER', 'KNOWN_TRANSFORMS', 'MOCK_GENES']

# %% ../nbs/00_static.ipynb 3
LOG_SAFETY = 1e-8

SSR_CHI2TEST = 'ssr_chi2test'
LAG_ORDER = 1

DEFAULT_TEST = SSR_CHI2TEST

DEFAULT_GRANGER_CAUSALITY_TESTS = dict(
    test      =  DEFAULT_TEST,
    lag_order =  LAG_ORDER,
    max_lag   = (LAG_ORDER, ) 
)


LOG2_FOLD = 'log2_fold'
SIGNED_CORRELATION= 'signed_correlation'
STANDARD_SCALER = 'standard_scaler'

KNOWN_TRANSFORMS = [LOG2_FOLD, SIGNED_CORRELATION, STANDARD_SCALER]

MOCK_GENES = 'wasf colq gpr1 chrm3 lmod2 tek kank3 oca2 taz map4k1'.split()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_op.ipynb.

# %% auto 0
__all__ = ['GrangerCausality']

# %% ../nbs/03_op.ipynb 3
from sklearn.base import BaseEstimator
import numpy as np, pandas as pd

import seaborn as sns

from dataclasses import dataclass, field
from typing import Optional, ClassVar

# %% ../nbs/03_op.ipynb 4
from atyp import SeriesLike, DataFrame

# %% ../nbs/03_op.ipynb 5
from gnr.static import (
    LAG_ORDER, DEFAULT_TEST,

    LOG2_FOLD, SIGNED_CORRELATION, STANDARD_SCALER, KNOWN_TRANSFORMS
)

from gnr.utils import (    
    apply_log2_fold, apply_signed_correlation, apply_standard_scaler,
    _prep_args_for_granger_causality_tests,
    _prep_vars_for_granger_causality_tests
)

from gnr.fn import (
    shift_trajectories, get_pval_from_granger_causality_tests, 
    grangers_causation_matrix, calculate_granger_causation
)

# %% ../nbs/03_op.ipynb 6
class GrangerCausality(BaseEstimator):
    '''
    Computes Granger Causality
    Check Granger Causality of all possible combinations of the Time series.

    The rows are the response variable, columns are predictors. The values in the table 
    are the P-Values. P-Values lesser than the significance level (0.05), implies 
    the Null Hypothesis that the coefficients of the corresponding past values is 
    zero, that is, the X does not cause Y can be rejected.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (features), and 
        columns are predictors (expression).
    
    x_vars : SeriesLike, optional
        A subset of response variable (features) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.


    y_vars : SeriesLike, optional
        A subset of response variable (features) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.

    shift : int, default=10
        number to shift `df`'s values by

    test : str, default='ssr_chi2test'
        the kind of statistical test to use
    
    lag_order : int, default=1
        how long to lag
    
    max_lag : tuple, optional
        if `None` coerced to `(1, )`

    n_jobs : int, default=-1
        number of cpu threads to use during calculation
        
    Returns
    -------
    df_res : pd.DataFrame    
        Pandas DataFrame with shape `(len(x_vars), len(y_vars))` containing the
        minimum p-value from Granger's Causation Tests

    Notes
    -----
    If `x_var = ['A', B']` and `y_var = ['C', 'D]', the Granger's Causality matrix we
    return has shape:
    ```
        | C_y | D_y |
        -------------
    | A_x |      |     |
    --------------------
    | B_x |      |     |

    ```

    See Also
    --------
    GrangerCausality._prep_args_for_granger_causality_tests :
        validates input arguments

    GrangerCausality._prep_vars_for_granger_causality_tests :
        validates input variables
    '''
    _LOG2: ClassVar[str] = LOG2_FOLD
    _SCOR: ClassVar[str] = SIGNED_CORRELATION
    _STDN: ClassVar[str] = STANDARD_SCALER
    KNOWN_TRANSFORMS: ClassVar[str] = KNOWN_TRANSFORMS


    def __init__(
        self,
        x_vars: Optional[SeriesLike] = None, 
        y_vars: Optional[SeriesLike] = None,
        shift: Optional[int] = 10,
        test: Optional[str] = DEFAULT_TEST,
        lag_order: Optional[int] = LAG_ORDER,
        max_lag: Optional[tuple] = (LAG_ORDER, ),
        n_jobs: Optional[int] = -1
    ):        
        
    

        test, lag_order, max_lag = _prep_args_for_granger_causality_tests(test, lag_order, max_lag)

        self.x_vars = x_vars
        self.y_vars = y_vars
        self.shift = shift
        self.test = test
        self.max_lag = max_lag
        self.lag_order = lag_order
        self.n_jobs = n_jobs

        # original DataFrame
        self.df_org = None

        # results DataFrame
        self.df_res = None
        
        # scalar if transforming
        self.scaler = None

    
    def process_fit_params(self, **fit_params) -> dict:
        '''
        fit_params : dict, optional

        NOTE: you can use either of the following. The first is more explicit, the second is more concise.
        ```
            {
                "apply" : str, default=None | "log2_fold" | "signed_correlation" | "standard_scaler"
            }

            {
                "use_cached" : bool, default=False
                "log2_fold" : bool, default=False
                "signed_correlation" : bool, default=False
                "standard_scaler" : bool, default=False
            }
        ```
        if `apply` is not `None`, then the following apply:
        - `log2_fold` : will apply log2 fold change to `df_res`
        - `signed_correlation` : will apply signed correlation to `df_res`
        - `standard_scaler` : will apply standard scaler to `df_org` and `df_res`
        '''
        apply = fit_params.get('apply', None)
        if apply is not None and apply not in self.KNOWN_TRANSFORMS:
            if isinstance(apply, str):
                for app in apply.split(' '):
                    if app not in self.KNOWN_TRANSFORMS:
                        continue
                    fit_params[app] = True
        return fit_params



    def fit(self, X):
        '''
        Update internal parameters prior to computation in `transform`
        '''
        test, lag_order, max_lag = _prep_args_for_granger_causality_tests(self.test, self.lag_order, self.max_lag)
        X, x_vars, y_vars = _prep_vars_for_granger_causality_tests(X, self.x_vars, self.y_vars, do_safety_check=True)
        self.test = test
        self.lag_order = lag_order
        self.max_lag = max_lag
        self.x_vars = x_vars
        self.y_vars = y_vars
        return self
    
    def transform(self, X: DataFrame, y = None, **fit_params):   
        # Save input DataFrame
        self.df_org = X.copy()
        fit_params = self.process_fit_params(**fit_params)

        # Apply standard scaler
        stdn = fit_params.get(self._STDN, False)
        if stdn:
            self.apply_standard_scaler(X)

        # Use saved results
        use_cached = fit_params.get('use_cached', False)
        if use_cached and self.df_res is not None:
            df_res = self.df_res.copy()

        else:
            df_res = calculate_granger_causation(
                X, x_vars=self.x_vars, y_vars=self.y_vars,
                shift=self.shift, test=self.test, 
                lag_order=self.lag_order, max_lag=self.max_lag,    
                n_jobs=self.n_jobs
            )

        # Store results
        self.df_res = df_res.copy()

        log2 = fit_params.get(self._LOG2, False)
        if log2:
            df_res = apply_log2_fold(df_res)

        scor = fit_params.get(self._SCOR, False)
        if scor:
            df_res = apply_signed_correlation(X, df_res)        

        return df_res
    
    def fit_transform(self, X: DataFrame, y = None, **fit_params):
        df_res = self.fit(X).transform(X, y = y, **fit_params)
        return df_res
    
    def apply_log2_fold(self, df: DataFrame) -> DataFrame:
        return apply_log2_fold(df)

    def apply_signed_correlation(self, df_tseries: DataFrame, df_granger: DataFrame) -> pd.DataFrame:
        return apply_signed_correlation(df_tseries, df_granger)
    
    def apply_standard_scaler(self, df: DataFrame) -> DataFrame:
        df_res, scaler = apply_standard_scaler(df, return_scaler=True)
        self.scaler = scaler
        return df_res
    
    def invert_scaler(self, df: DataFrame) -> DataFrame:
        if self.scaler is None:
            raise ValueError
        return self.scaler.inverse_transform(df)
    
    def plot_df_org(self, show_all_yticks: bool = True, **kwargs):
        '''
        Returns
        -------
        fig : matplotlib.pyplot.Figure
        ax : matplotlib.pyplot.Axis
        ClusterMap : seaborn.ClusterMap
        '''
        options = dict(cmap='inferno', robust=True, col_cluster=False, yticklabels=show_all_yticks)
        options.update(kwargs)
        cstrmp = sns.clustermap(self.df_org, **options)
        return cstrmp
    
    def plot_df_res(self, show_all_yticks: bool = True, **kwargs):
        '''
        Returns
        -------
        fig : matplotlib.pyplot.Figure
        ax : matplotlib.pyplot.Axis
        ClusterMap : seaborn.ClusterMap
        '''           
        options = dict(yticklabels=show_all_yticks)
        options.update(kwargs)
        cstrmp = sns.clustermap(self.df_res, **options)
        return cstrmp




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_fn.ipynb.

# %% auto 0
__all__ = ['shift_trajectories', 'get_pval_from_granger_causality_tests', 'grangers_causation_matrix',
           'calculate_granger_causation']

# %% ../nbs/02_fn.ipynb 3
import warnings
import pandas as pd, numpy as np

from joblib import Parallel, delayed
from typing import Optional

from statsmodels.tsa.stattools import grangercausalitytests

# %% ../nbs/02_fn.ipynb 4
from atyp import SeriesLike, DataFrame

# %% ../nbs/02_fn.ipynb 5
from gnr.static import (
    LOG_SAFETY, SSR_CHI2TEST, LAG_ORDER,
    DEFAULT_TEST, DEFAULT_GRANGER_CAUSALITY_TESTS
)

from gnr.utils import (
    _add_suffix, make_stable, add_stability, add_non_const, 
    apply_log2_fold, apply_signed_correlation, apply_standard_scaler,
    _prep_args_for_granger_causality_tests,
    _prep_vars_for_granger_causality_tests
)

# %% ../nbs/02_fn.ipynb 7
def shift_trajectories(
    df: DataFrame, 
    shift: Optional[int] = 10, 
    copy: Optional[bool] = True
) -> DataFrame:
    '''
    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (genes), and 
        columns are predictors (expression).
    
    shift : int, default=10
        number to shift `df`'s values by

    copy : bool, default=True
        Whether or not to copy input `df`

    Returns
    -------
    df_trj : pd.DataFrame
        Pandas DataFrame of `df - df_shift`
    '''
    df_trj = (df.copy() if copy else df)
    df_trj = df_trj.T[::shift]
    df_trj = df_trj - df_trj.shift(1)
    df_trj = df_trj.dropna()
    return df_trj


# %% ../nbs/02_fn.ipynb 8
def get_pval_from_granger_causality_tests(
    df: DataFrame,
    test: Optional[str] = DEFAULT_TEST, 
    lag_order: Optional[int] = LAG_ORDER,
    max_lag: Optional[tuple] = (LAG_ORDER, ), 
) -> float:
    '''
    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (genes), and 
        columns are predictors (expression).

    test : str, default='ssr_chi2test'
        the kind of statistical test to use
    
    lag_order : int, default=1
        how long to lag
    
    max_lag : tuple, optional
        if `None` coerced to `(1, )`

    Returns
    -------
    min_p_value : float
        minimum p-value of Granger Causality Tests

    '''
    test, lag_order, max_lag = _prep_args_for_granger_causality_tests(test, lag_order, max_lag)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        test_result = grangercausalitytests(df, maxlag=max_lag, verbose=False)
    p_values = [test_result[i][0][test][1] for i in max_lag]
    min_p_value = np.min(p_values)
    return min_p_value

# %% ../nbs/02_fn.ipynb 10
def grangers_causation_matrix(
    df: DataFrame, 
    x_vars: Optional[SeriesLike] = None, 
    y_vars: Optional[SeriesLike] = None, 
    test: Optional[str] = DEFAULT_TEST, 
    lag_order: Optional[int] = LAG_ORDER,
    max_lag: Optional[tuple] = (LAG_ORDER, ),    
    n_jobs: Optional[int] = -1
) -> DataFrame:
    '''
    Computes Granger Causality

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (genes), and 
        columns are predictors (expression).
    
    x_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.

    y_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.

    test : str, default='ssr_chi2test'
        the kind of statistical test to use
    
    lag_order : int, default=1
        how long to lag
    
    max_lag : tuple, optional
        if `None` coerced to `(1, )`

    n_jobs : int, default=-1
        number of cpu threads to use during calculation
        
    Returns
    -------
    df_res : pd.DataFrame    
        Pandas DataFrame with shape `(len(x_vars), len(y_vars))` containing the
        minimum p-value from Granger's Causation Tests

    Notes
    -----
    If `x_var = ['A', B']` and `y_var = ['C', 'D]', the Granger's Causality matrix we
    return has shape:
    ```
           | C_y | D_y |
           -------------
    | A_x |      |     |
    --------------------
    | B_x |      |     |

    ```
    '''
    n_jobs = -1 if n_jobs is None else n_jobs
    
    df, x_vars, y_vars       = _prep_vars_for_granger_causality_tests(df, x_vars, y_vars)
    test, lag_order, max_lag = _prep_args_for_granger_causality_tests(test, lag_order, max_lag)

    delayed_funcs = [
        delayed(get_pval_from_granger_causality_tests)(
            df[[x_var, y_var]], test, lag_order, max_lag
        ) for y_var in y_vars for x_var in x_vars
    ]

    parallel_out = Parallel(n_jobs=n_jobs)(delayed_funcs)
    
    df_res = pd.DataFrame(
        np.array(parallel_out).reshape((len(x_vars), len(y_vars))),
        index=x_vars, columns=y_vars,        
    )
    df_res.index = _add_suffix(x_vars, 'x')
    df_res.columns = _add_suffix(y_vars, 'y')
    return df_res

# %% ../nbs/02_fn.ipynb 11
def calculate_granger_causation(
    df: DataFrame, 
    x_vars: Optional[SeriesLike] = None, 
    y_vars: Optional[SeriesLike] = None, 
    shift: Optional[int] = 10,
    test: Optional[str] = DEFAULT_TEST, 
    lag_order: Optional[int] = LAG_ORDER,
    max_lag: Optional[tuple] = (LAG_ORDER, ),    
    n_jobs: Optional[int] = -1
) -> DataFrame:     
    '''
    Computes Granger Causality

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (genes), and 
        columns are predictors (expression).
    
    x_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.


    y_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.

    shift : int, default=10
        number to shift `df`'s values by

    test : str, default='ssr_chi2test'
        the kind of statistical test to use
    
    lag_order : int, default=1
        how long to lag
    
    max_lag : tuple, optional
        if `None` coerced to `(1, )`

    n_jobs : int, default=-1
        number of cpu threads to use during calculation
        
    Returns
    -------
    df_res : pd.DataFrame    
        Pandas DataFrame with shape `(len(x_vars), len(y_vars))` containing the
        minimum p-value from Granger's Causation Tests

    Notes
    -----
    If `x_var = ['A', B']` and `y_var = ['C', 'D]', the Granger's Causality matrix we
    return has shape:
    ```
           | C_y | D_y |
           -------------
    | A_x |      |     |
    --------------------
    | B_x |      |     |

    ```
    '''
    n_jobs = -1 if n_jobs is None else n_jobs
    
    df, x_vars, y_vars       = _prep_vars_for_granger_causality_tests(df, x_vars, y_vars)    
    test, lag_order, max_lag = _prep_args_for_granger_causality_tests(test, lag_order, max_lag)
    
    df_trj = shift_trajectories(df, shift)
        
    df_res = grangers_causation_matrix(
        df_trj, x_vars=x_vars, y_vars=y_vars,
        test=test, lag_order=lag_order, max_lag=max_lag,
        n_jobs=n_jobs
    )

    return df_res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_utils.ipynb.

# %% auto 0
__all__ = ['make_stable', 'add_stability', 'add_non_const', 'apply_log2_fold', 'apply_signed_correlation',
           'apply_standard_scaler', 'kpss_test']

# %% ../nbs/01_utils.ipynb 3
import pandas as pd, numpy as np
import seaborn as sns

from typing import List, Optional, Union

from sklearn.preprocessing import StandardScaler
from statsmodels.tsa.stattools import kpss

# %% ../nbs/01_utils.ipynb 5
from atyp import SeriesLike, DataFrame, NPArray

# %% ../nbs/01_utils.ipynb 6
from gnr.static import (
    LOG_SAFETY, SSR_CHI2TEST, LAG_ORDER,
    DEFAULT_TEST, DEFAULT_GRANGER_CAUSALITY_TESTS,
    MOCK_GENES
)

# %% ../nbs/01_utils.ipynb 8
def _add_suffix(arr, suffix:str):
    return list(map(lambda e: f'{e}_{suffix}', arr))

def make_stable(
    df: DataFrame, 
    copy: Optional[bool] = True,
    add_perm: Optional[bool] = True,
    use_rand: Optional[bool] = False,
    no_const: Optional[bool] = True
) -> DataFrame:    
    df_safe = (df.copy() if copy else df)

    if add_perm:
        df_safe = df_safe + LOG_SAFETY
    
    # NOTE: to ensure that no column in df is constant
    if use_rand or no_const:
        rng_mat = np.random.randn(*df.shape) * LOG_SAFETY
        df_safe = df_safe + rng_mat

    return df_safe
    
def add_stability(df):
    return make_stable(df, copy=False, add_perm=True)

def add_non_const(df):
    return make_stable(df, copy=False, add_perm=False, use_rand=True)
    
def apply_log2_fold(df_granger: DataFrame):
    return -np.log2(df_granger + (2 ** -20))    

def apply_signed_correlation(df_tseries: DataFrame, df_granger: DataFrame) -> DataFrame:
    var_names = df_tseries.index.values
    signed = -np.sign(df_tseries.T.corr()) * np.log(add_stability(df_granger.values))
    df_res = pd.DataFrame(signed, index=var_names, columns=var_names)
    return df_res

def apply_standard_scaler(df: DataFrame, return_scaler: Optional[bool] = False) -> DataFrame:
    scaler = StandardScaler()
    scaler = scaler.fit(df.T)
    df_res = pd.DataFrame(scaler.transform(df.T)).T
    df_res.index = df.index
    if return_scaler:
        df_res, scaler
    return df_res

# %% ../nbs/01_utils.ipynb 10
def kpss_test(series, print: Optional[bool] = False, **kwargs) -> None:
    # KPSS Null hypothesis: there is a no unit root, meaning series is stationary
    statistic, p_value, n_lags, critical_values = kpss(series, **kwargs)

    # Format Output    
    result_str = ''
    result_str += f'KPSS Statistic: {statistic}\n'
    result_str += f'p-value: {p_value}\n'
    result_str += f'num lags: {n_lags}\n'
    result_str += f'Critial Values:\n'    
    for key, value in critical_values.items():
        result_str += f'\t{key}:{value}\n'
    result_str += f'Result: The series is {"not " if p_value < 0.05 else ""}stationary'

    if print:
        print(result_str)
    return

# %% ../nbs/01_utils.ipynb 12
def _prep_args_for_granger_causality_tests(
    test: Optional[str] = DEFAULT_TEST, 
    lag_order: Optional[int] = LAG_ORDER,
    max_lag: Optional[tuple] = (LAG_ORDER, ),    
):
    '''
    Parameters
    ----------
    test : str, default='ssr_chi2test'
        the kind of statistical test to use
    
    lag_order : int, default=1
        how long to lag
    
    max_lag : tuple, optional
        if `None` coerced to `(1, )`
    
    Returns
    -------
    test : str
        `'ssr_chi2test'` if `test` was `None`
        
    lag_order : int
        `1` if `lag_order` was `None`

    max_lag : tuple
        `(1, )` if `max_lag` was `None`

    Notes
    -----
    lag_order
        `lag_order = 1` # becuase we got this value before. We are not suppose to add 1 to it
    
    max_lag
        `max_lag = (lag_order, )`
    '''
    if test is None:
        test = SSR_CHI2TEST

    if lag_order is None:
        lag_order = LAG_ORDER

    if max_lag is None:
        max_lag = (lag_order,)
    
    return test, lag_order, max_lag

def _prep_vars_for_granger_causality_tests(
    df: DataFrame, 
    x_vars: Optional[SeriesLike] = None, 
    y_vars: Optional[SeriesLike] = None,
    do_safety_check: Optional[bool] = True     
):
    '''
    Ensures `df` is safe for Granger Causality tests and that 
    `x_vars` and `y_vars` are not `None`.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame where rows are the response variable (genes), and 
        columns are predictors (expression).
    
    x_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.


    y_vars : SeriesLike, optional
        A subset of response variable (genes) to compute granger's causality test with. 
        If not provided, defaults to `df.index.values` i.e. all rows in `df`.

    do_safety_check : bool, default=True
        Whether or not to check if there is a constant column in `df`.

        
    Returns
    -------
    df : pd.DataFrame
        input `df` possibly with small random noise added to prevent constant columns

    x_vars : SeriesLike
        input `x_vars` or `df.index.values` if `x_vars` was `None`

    y_vars : SeriesLike
        input `y_vars` or `df.index.values` if `y_vars` was `None`

    Notes
    -----
    If `x_var = ['A', B']` and `y_var = ['C', 'D]', then when calling Granger's 
    Causality the matrix returned has shape:
    ```
        | C | D |
        ---------
    | A |   |   |
    -------------
    | B |   |   |

    ```
    '''
    if do_safety_check:
        df = add_non_const(df)

    if x_vars is None:
        x_vars = df.index.values

    if y_vars is None:
        y_vars = df.index.values
    
    return df, x_vars, y_vars



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_mock.ipynb.

# %% auto 0
__all__ = ['make_dummy_genes', 'make_dummy_cells', 'make_mock_genes_x_tbins', 'make_mock_trajectories', 'df_trj_to_cells_x_tbins',
           'df_trj_to_genes_x_tbins']

# %% ../nbs/04_mock.ipynb 3
import pandas as pd, numpy as np
from typing import List, Optional, Union

# %% ../nbs/04_mock.ipynb 4
from atyp import SeriesLike, DataFrame, NPArray

# %% ../nbs/04_mock.ipynb 5
from .static import (MOCK_GENES)

# %% ../nbs/04_mock.ipynb 6
def make_dummy_genes(n:int):
    return [f'gene_{i}' for i in range(1, n+1)]

def make_dummy_cells(n:int):
    return [f'cell{i}' for i in range(1, n+1)]

# %% ../nbs/04_mock.ipynb 7
def make_mock_genes_x_tbins(
    genes: Optional[Union[int, List[str]]] = MOCK_GENES,
    tbins: int = 100,
) -> DataFrame:
    
    if isinstance(genes, int):
        genes = make_dummy_genes(genes)        
    n_genes = len(genes)

    forward = lambda : np.sort(np.random.randint(0, n_genes, (tbins)))
    reverse = lambda : forward()[::-1]

    timeseries = [forward() if i % 2 == 0 else reverse() for i in range(n_genes)]
    df_trj = pd.DataFrame(timeseries, index=genes)
    return df_trj

# %% ../nbs/04_mock.ipynb 9
def make_mock_trajectories(
    tbins: int = 100,
    cells: int = 50,
    genes: int = 10,
) -> NPArray:
    df = pd.DataFrame()
    for i in range(cells):
        df_gxt = make_mock_genes_x_tbins(genes, tbins)
        df = pd.concat([df, df_gxt])
    res = df.values.reshape(tbins, cells, genes)
    return res

# %% ../nbs/04_mock.ipynb 11
def df_trj_to_cells_x_tbins(trajectories:NPArray, agg_fn=np.mean) -> NPArray:
    '''
    Transpose and aggregate trajectories matrix (timebinse, cells, gene)
    to produce (cells, timebins)
    ''' 
    return agg_fn(np.transpose(trajectories, (1, 0, 2)), axis=2)

def df_trj_to_genes_x_tbins(trajectories:NPArray, agg_fn=np.mean) -> NPArray:
    '''
    Transpose and aggregate trajectories matrix (timebinse, cells, gene)
    to produce (genes, timebins)
    '''
    return agg_fn(np.transpose(trajectories, (2, 0, 1)), axis=2)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/gnr',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/gnr',
                'lib_path': 'gnr'},
  'syms': { 'gnr.fn': { 'gnr.fn.calculate_granger_causation': ('fn.html#calculate_granger_causation', 'gnr/fn.py'),
                        'gnr.fn.get_pval_from_granger_causality_tests': ('fn.html#get_pval_from_granger_causality_tests', 'gnr/fn.py'),
                        'gnr.fn.grangers_causation_matrix': ('fn.html#grangers_causation_matrix', 'gnr/fn.py'),
                        'gnr.fn.shift_trajectories': ('fn.html#shift_trajectories', 'gnr/fn.py')},
            'gnr.mock': { 'gnr.mock.df_trj_to_cells_x_tbins': ('mock.html#df_trj_to_cells_x_tbins', 'gnr/mock.py'),
                          'gnr.mock.df_trj_to_genes_x_tbins': ('mock.html#df_trj_to_genes_x_tbins', 'gnr/mock.py'),
                          'gnr.mock.make_dummy_cells': ('mock.html#make_dummy_cells', 'gnr/mock.py'),
                          'gnr.mock.make_dummy_genes': ('mock.html#make_dummy_genes', 'gnr/mock.py'),
                          'gnr.mock.make_mock_genes_x_tbins': ('mock.html#make_mock_genes_x_tbins', 'gnr/mock.py'),
                          'gnr.mock.make_mock_trajectories': ('mock.html#make_mock_trajectories', 'gnr/mock.py')},
            'gnr.op': { 'gnr.op.GrangerCausality': ('op.html#grangercausality', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.__init__': ('op.html#grangercausality.__init__', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.apply_log2_fold': ('op.html#grangercausality.apply_log2_fold', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.apply_signed_correlation': ( 'op.html#grangercausality.apply_signed_correlation',
                                                                              'gnr/op.py'),
                        'gnr.op.GrangerCausality.apply_standard_scaler': ('op.html#grangercausality.apply_standard_scaler', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.fit': ('op.html#grangercausality.fit', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.fit_transform': ('op.html#grangercausality.fit_transform', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.invert_scaler': ('op.html#grangercausality.invert_scaler', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.plot_df_org': ('op.html#grangercausality.plot_df_org', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.plot_df_res': ('op.html#grangercausality.plot_df_res', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.process_fit_params': ('op.html#grangercausality.process_fit_params', 'gnr/op.py'),
                        'gnr.op.GrangerCausality.transform': ('op.html#grangercausality.transform', 'gnr/op.py')},
            'gnr.static': {},
            'gnr.utils': { 'gnr.utils._add_suffix': ('utils.html#_add_suffix', 'gnr/utils.py'),
                           'gnr.utils._prep_args_for_granger_causality_tests': ( 'utils.html#_prep_args_for_granger_causality_tests',
                                                                                 'gnr/utils.py'),
                           'gnr.utils._prep_vars_for_granger_causality_tests': ( 'utils.html#_prep_vars_for_granger_causality_tests',
                                                                                 'gnr/utils.py'),
                           'gnr.utils.add_non_const': ('utils.html#add_non_const', 'gnr/utils.py'),
                           'gnr.utils.add_stability': ('utils.html#add_stability', 'gnr/utils.py'),
                           'gnr.utils.apply_log2_fold': ('utils.html#apply_log2_fold', 'gnr/utils.py'),
                           'gnr.utils.apply_signed_correlation': ('utils.html#apply_signed_correlation', 'gnr/utils.py'),
                           'gnr.utils.apply_standard_scaler': ('utils.html#apply_standard_scaler', 'gnr/utils.py'),
                           'gnr.utils.kpss_test': ('utils.html#kpss_test', 'gnr/utils.py'),
                           'gnr.utils.make_stable': ('utils.html#make_stable', 'gnr/utils.py')}}}
